import java.util.*;
import java.net.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class Client 
  implements IRKeypad.IRListener {

  public static class ClientReceiver implements Runnable {
    public DatagramPacket p_;
    Client c_;
    public ClientReceiver(Client c, DatagramPacket p) { c_ = c; p_ = p; }
    public void run() { c_.recvPacket(p_); }
  }

  private VFDUnit vfd_;
  private IRKeypad remote_ = null;
  private StreamController stream_;

  private DatagramSocket s_;
  private InetAddress serverIP_ = null;
  private int serverPort_ = 1069;
  private int localPort_  = -1;

  private int displaySize_ = VFDUnit.SIZE_NORMAL;
  private boolean showRemote_ = true;

  private long startTime_;

  private byte buf_[];
  private DatagramPacket pkt_;

  private static final byte devID_ = 1;
  private static final byte firmwareVers_ = 0x14;

  private static final String myVersion_ = "0.1.1pre1";

  private int debugLevel_ = 0;

  public static void main(String args[]) {
    Client c = new Client(args);
  }

  public Client(String args[]) {
    parseArgs(args);

    Util.setDebugLevel(debugLevel_);
    startTime_ = System.currentTimeMillis();
    
    s_ = null;
    try { 
      if (localPort_ > 0)
	s_ = new DatagramSocket(localPort_); 
      else
	s_ = new DatagramSocket(); 
    }
    catch (SocketException x) { Util.fail("Unable to allocate socket: " + x.getMessage()); }

    Listener l = new Listener(this);
    l.start();

    stream_ = new StreamController();

    vfd_ = new VFDUnit(displaySize_);

    JFrame frame = new JFrame("TripleFat " + myVersion_);
    frame.getContentPane().add(vfd_.getScreen(), BorderLayout.CENTER);
    frame.addWindowListener(
      new WindowAdapter() {
	  public void windowClosing(WindowEvent e) {
	      System.exit(0);
	  } } );
    frame.pack();
    frame.setVisible(true);
    
    if (showRemote_) {
      remote_ = new IRKeypad();
      remote_.addListener(this);
      
      JFrame frame2 = new JFrame("IR Remote");
      frame2.getContentPane().add(remote_, BorderLayout.CENTER);
      frame2.addWindowListener(
	new WindowAdapter() {
	  public void windowClosing(WindowEvent e) {
	    System.exit(0);
	  } } );
      frame2.pack();
      frame2.setVisible(true);
    }
    
    if (serverIP_ == null) {
      // Try to contact server with UDP broadcast.  This might fail
      // on FreeBSD and Digital Unix, because they don't enable the
      // socket's broadcast option by default.  See
      // e.g. http://www.geocrawler.com/mail/msg.php3?msg_id=808473&list=162
      System.out.println("Searching for server on network");
      try { serverIP_ = InetAddress.getByName("255.255.255.255"); }
      catch (UnknownHostException x) { Util.fail("Unable to get broadcast address"); }
      allocatePacket();
      sendDiscoveryRequest();
    }
    else {
      System.out.println("Using server " + serverIP_ + ":" + serverPort_);
      allocatePacket();
      sendHello();
    }
  }

  private void sendHello() {
    buf_[0] = 'h';
    buf_[1] = devID_;
    buf_[2] = firmwareVers_;
    for (int i = 3; i < 18; i++)
      buf_[i] = ' ';
    pkt_.setLength(18);
    try { s_.send(pkt_); }
    catch (IOException x) { 
      Util.dbg("Unable to send Hello packet: " + x.getMessage()); 
    }
  }

  private void sendDiscoveryRequest() {
    buf_[0] = 'd';
    buf_[1] = devID_;
    buf_[2] = firmwareVers_;
    for (int i = 3; i < 18; i++)
      buf_[i] = ' ';
    pkt_.setLength(18);
    try { s_.send(pkt_); }
    catch (IOException x) { 
      Util.dbg("Unable to send Discovery Request packet: " + x.getMessage()); 
    }
  }

  public void sendStreamRequest(int writePos, int readPos) {
    buf_[0] = 'r';
    buf_[1] = ' ';
    buf_[2] = (byte) (0xff & (writePos >> 8));
    buf_[3] = (byte) (0xff & writePos);
    buf_[4] = (byte) (0xff & (readPos >> 8));
    buf_[5] = (byte) (0xff & readPos);
    for (int i = 6; i < 18; i++)
      buf_[i] = ' ';
    pkt_.setLength(18);
    try { s_.send(pkt_); }
    catch (IOException x) { 
      Util.dbg("Unable to send Stream Request packet: " + x.getMessage()); 
    }
  }

  private void sendIRPacket(int ticks, int code) {
    buf_[0] = 'i';
    buf_[1] = 0x00;
    // net byte-order
    buf_[2] = (byte) (0xff & (ticks >> 24));
    buf_[3] = (byte) (0xff & (ticks >> 16));
    buf_[4] = (byte) (0xff & (ticks >> 8));
    buf_[5] = (byte) (0xff & ticks);
    buf_[6] = (byte) 0xff;
    buf_[7] = (byte) 0x16; // XXX
    buf_[8] = (byte) (0xff & (code >> 24));
    buf_[9] = (byte) (0xff & (code >> 16));
    buf_[10] = (byte) (0xff & (code >> 8));
    buf_[11] = (byte) (0xff & code);
    for (int i = 12; i < 17; i++)
      buf_[i] = ' ';
    pkt_.setLength(18);
    try { s_.send(pkt_); }
    catch (IOException x) { 
      Util.dbg("Unable to send IR packet: " + x.getMessage()); 
    }
  }

  public void recvPacket(DatagramPacket p) {
    int type = p.getLength() > 0 ? (0xff & p.getData()[p.getOffset()]) : -1;
    
    if (!(p.getLength() >= 18 || (type == 's' && p.getLength() >= 2))) {
      Util.dbg("Client.recvPacket: Ignoring short packet");
      Util.dbg(dumpPacket(p));
      return;
    }

    if      (type == 'D') handleDiscoveryPacket(p);
    else if (type == 'h') handleHelloPacket(p);
    else if (type == 'l') handleDisplayPacket(p);
    else if (type == 's') handleStreamControlPacket(p);
    else if (type == 'm') handleMPEGDataPacket(p);
    else if (type == '2') handleBusPacket(p);
    else
      Util.dbg("Client.recvPacket: unknown packet type 0x" + Integer.toHexString(type));
  }

  private void handleDiscoveryPacket(DatagramPacket p) {
    byte b[] = p.getData();
    int offset = p.getOffset();
    if (b[2 + offset] == 0 && b[3 + offset] == 0 && b[4 + offset] == 0 && b[5 + offset] == 0) {
      // extract server IP/port from UDP header
      serverIP_ = p.getAddress();
      serverPort_ = p.getPort();
    }
    else {
      // extract server IP/port from packet contents
      String addr = b[2 + offset] + "." + b[3 + offset] + "." + b[4 + offset] + "." + b[5 + offset];
      InetAddress ip = null;
      try { ip = InetAddress.getByName(addr); }
      catch (UnknownHostException x) { 
	Util.dbg("Client.handleDiscoveryPacket: couldn't get server IP from packet contents");
	return;
      }
      serverIP_ = ip;
      serverPort_ = ((0xff & b[6 + offset]) << 8) | (0xff & b[7 + offset]);
    }
    Util.dbg("Setting server address to " + serverIP_ + ":" + serverPort_);
  }

  
  private void handleHelloPacket(DatagramPacket p) {
    // switch to this server
    serverIP_ = p.getAddress();
    serverPort_ = p.getPort();
    pkt_.setAddress(serverIP_);
    pkt_.setPort(serverPort_);

    sendHello();    
  }

  private void handleDisplayPacket(DatagramPacket p) {
    vfd_.handleCodes(p.getData(), 18 + p.getOffset(), p.getLength() - 18);
  }

  private void handleStreamControlPacket(DatagramPacket p) {
    byte b[] = p.getData();
    int offset = p.getOffset();
    
    int code  = 0xff & b[1 + offset];
    int delay = p.getLength() > 2 ? (0xff & b[2 + offset]) : 0;
    
    if (delay == 0)
      handleStreamCommand(code);
    else {
      class DelayedStreamAction implements ActionListener {
	Client c_;
	int code_;
	DelayedStreamAction(Client c, int code) { c_ = c; code_ = code; }
	public void actionPerformed(ActionEvent e) { c_.handleStreamCommand(code_); }
      }
      javax.swing.Timer t = new javax.swing.Timer(delay, new DelayedStreamAction(this, code));
      t.setRepeats(false);
      t.start();
    }
  }

  private void handleMPEGDataPacket(DatagramPacket p) {
    if (((p.getLength() - 18) & 1) != 0) {
      Util.dbg("Client.handleMPEGDataPacket: ignoring packet with odd number of data bytes");
      return;
    }

    byte b[] = p.getData();
    int offset = p.getOffset();
    int control = 0xff & b[offset + 1];
    int write_ptr = ((0xff & b[offset + 6]) << 8) | (0xff & b[offset + 7]);
    int seq = ((0xff & b[offset + 10]) << 8) | (0xff & b[offset + 11]);

    stream_.write(write_ptr, b, offset + 18, p.getLength() - 18);

    if (control == 0)      stream_.resume(); 
    else if (control == 1) stream_.pause();
    else if (control == 3) stream_.reset();

    // send ACK
    buf_[0] = 'a';
    buf_[1] = buf_[2] = buf_[3] = buf_[4] = buf_[5] = ' ';
    int w = stream_.getWPtr();
    int r = stream_.getRPtr();
    buf_[6] = (byte) (0xff & (w >> 8));
    buf_[7] = (byte) (0xff & w);
    buf_[8] = (byte) (0xff & (r >> 8));
    buf_[9] = (byte) (0xff & r);
    buf_[10] = (byte) (0xff & (seq >> 8));
    buf_[11] = (byte) (0xff & seq);
    for (int i = 12; i < 18; i++)
      buf_[i] = ' ';
    pkt_.setLength(18);

    try { s_.send(pkt_); }
    catch (IOException x) { 
      Util.dbg("Unable to send MPEG Data ACK packet: " + x.getMessage()); 
    }
  }

  private void handleBusPacket(DatagramPacket p) {
    Util.dbg("i2c pkt: " + dumpPacket(p));
    // send ACK
    buf_[0]  = '2';
    for (int i = 1; i < 18; i++)
      buf_[i] = ' ';
    buf_[18] = 'a';

    pkt_.setLength(19);
    try { s_.send(pkt_); }
    catch (IOException x) { 
      Util.dbg("Unable to send i2c ACK packet: " + x.getMessage()); 
    }
  }

  static final private int STREAM_RESET  = 1;
  static final private int STREAM_PAUSE  = 2;
  static final private int STREAM_RESUME = 4;
  static final private int STREAM_USAGE  = 8;

  private void handleStreamCommand(int code) {
    if ((code & STREAM_RESET) != 0)
      stream_.reset();
    if ((code & STREAM_PAUSE) != 0)
      stream_.pause();
    if ((code & STREAM_RESUME) != 0)
      stream_.resume();
    if ((code & STREAM_USAGE) != 0)
      stream_.showBufferUsage();
  }

  // from IRListener
  public void keyPressed(int code) {
    int ticks = getTicks();
    sendIRPacket(ticks, code);
  }

  public DatagramSocket getSocket() { 
    return s_; 
  }

  private int getTicks() {
    // 1 sec = 625,000 ticks
    long msecs = System.currentTimeMillis() - startTime_;
    return (int) msecs * 625;
  }

  private void allocatePacket() {
    buf_ = new byte[1600];
    pkt_ = new DatagramPacket(buf_, buf_.length, serverIP_, serverPort_);
  }
  
  private String dumpBytes(byte[] b) { return dumpBytes(b, 0, b.length); }
  private String dumpBytes(byte[] b, int offset, int n) {
    String s = "";
    for (int i = offset; i < b.length && i < n + offset; i++) 
      s += Integer.toHexString(0xff & b[i]) + " ";
    return s;
  }

  private String dumpPacket(DatagramPacket p) {
    return dumpBytes(p.getData(), p.getOffset(), p.getLength());
  }

  private void parseArgs(String args[]) {
    int n = args.length;
    for (int i = 0; i < n; i++) {
      String s = args[i].toLowerCase();
      if (s.equals("--help") || s.equals("-h")) {
	usage();
	System.exit(0);
      }
      else if (s.equals("--version") || s.equals("-v")) {
	printVersion();
	System.exit(0);
      }
      else if (s.equals("--server") || s.equals("-s")) {
	if (i + 1 >= n) usage();
	i++;
	try { serverIP_ = InetAddress.getByName(args[i]); }
	catch (UnknownHostException x) { Util.failQuiet("Unable to get server IP address for " + args[i]); }
      }
      else if (s.equals("--port") || s.equals("-p")) {
	if (i + 1 >= n) usage();
	i++;
	try { serverPort_ = Integer.parseInt(args[i]); }
	catch (NumberFormatException x) { Util.failQuiet("Bad server port number " + args[i]); }
      }
      else if (s.equals("--local-port") || s.equals("-l")) {
	if (i + 1 >= n) usage();
	i++;
	try { localPort_ = Integer.parseInt(args[i]); }
	catch (NumberFormatException x) { Util.failQuiet("Bad local port number " + args[i]); }
      }
      else if (s.equals("--small")) 
	displaySize_ = VFDUnit.SIZE_SMALL;
      else if (s.equals("--tiny")) 
	displaySize_ = VFDUnit.SIZE_TINY;
      else if (s.equals("--no-remote") || s.equals("-n")) 
	showRemote_ = false;
      else if (s.equals("--debug") || s.equals("-d")) {
	if (i + 1 >= n) usage();
	i++;
	try { debugLevel_ = Integer.parseInt(args[i]); }
	catch (NumberFormatException x) { Util.failQuiet("Bad debug level number " + args[i]); }
      }
      else {
	usage();
	System.exit(-1);
      }
    }
  }

  private void usage() {
    System.out.println("`TripleFat' is a graphical emulator for the SliMP3 hardware mp3 player.");
    System.out.println("");
    System.out.println("Options:");
    System.out.println("  -h, --help               Print this message and exit.");
    System.out.println("  -v, --version            Print version information and exit.");
    System.out.println("  -s, --server ADDR        Use server at IP address ADDR.");
    System.out.println("  -p, --port PORT          Use server UDP port PORT.");
    System.out.println("  -l, --local-port PORT    Use local UDP port PORT");
    System.out.println("                           (so that the server remembers this player).");
    System.out.println("  --small                  Use small display screen.");
    System.out.println("  --tiny                   Use tiny display screen.");
    System.out.println("  --no-remote              Don't display remote control panel.");
    System.out.println("  -d, --debug N            Set debugging output level to N.");
  }

  private void printVersion() {
    System.out.println("TripleFat " + myVersion_);
    System.out.println("Copyright (C) 2003 Douglas S. J. De Couto");
    System.out.println("This is free software; see the source for copying conditions.");
    System.out.println("There is NO warranty, not even for merchantability or fitness for a");
    System.out.println("particular purpose.");
  }

}
