import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class IRKeypad 
  extends JPanel 
  implements ActionListener {
  
  private static class ButtonInfo {
    int code;
    String name;
    int x;
    int y;
    ButtonInfo(String n, int c, int x_, int y_)
      { code = c; name = n; x = x_; y = y_; }
  }

  private Set buttons_;  // set of ButtonInfo
  private Map commands_; // IR code -> button

  public IRKeypad() {
    super();
    
    listeners_ = new Vector();
    buttons_ = new HashSet();
    commands_ = new HashMap();

    addCommands(buttons_);

    // assumes there is a button for every place in the grid
    setLayout(new GridLayout(11, 3));
    // applyComponentOrientation(ComponentOrientation.LEFT_TO_RIGHT);
    
    for (int y = 0; y < 11; y++) {
      for (int x = 0; x < 3; x++) {
	ButtonInfo bi = getButtonForXY(x, y);
	if (bi != null) {
	  JButton b = new JButton(bi.name);
	  b.addActionListener(this);
	  add(b);
	  commands_.put(bi.name, new Integer(bi.code));
	}
	else {
	  add(new JPanel());
	}
      }
    }
  }

  public static interface IRListener {
    public void keyPressed(int code);
  }

  private Vector listeners_;
  public void addListener(IRListener l) {
    listeners_.add(l);
  }
    
  private void notifyListeners(int code) {
    for (int i = 0; i < listeners_.size(); i++)
      ((IRListener) listeners_.get(i)).keyPressed(code);
  }

  public void actionPerformed(ActionEvent e) { 
    String cmdName = e.getActionCommand();
    Integer i = (Integer) commands_.get(cmdName);
    Util.assert(i != null);
    notifyListeners(i.intValue());      
  }

  private ButtonInfo getButtonForXY(int x, int y) {
    for (Iterator i = buttons_.iterator(); i.hasNext(); ) {
      ButtonInfo bi = (ButtonInfo) i.next();
      if (bi.x == x && bi.y == y)
	return bi;
    }
    return null;
  }

  private void addCommands(Set s) {
    s.add(new ButtonInfo("1"            , 0x7689f00f, 0, 5));
    s.add(new ButtonInfo("2"	        , 0x768908f7, 1, 5));
    s.add(new ButtonInfo("3"	        , 0x76898877, 2, 5));
    s.add(new ButtonInfo("4"	        , 0x768948b7, 0, 6));
    s.add(new ButtonInfo("5"	        , 0x7689c837, 1, 6));
    s.add(new ButtonInfo("6"	        , 0x768928d7, 2, 6));
    s.add(new ButtonInfo("7"	        , 0x7689a857, 0, 7));
    s.add(new ButtonInfo("8"	        , 0x76896897, 1, 7));
    s.add(new ButtonInfo("9"	        , 0x7689e817, 2, 7));
    s.add(new ButtonInfo("0"	        , 0x76899867, 1, 8));
    s.add(new ButtonInfo("Now Playing"  , 0x76897887, 0, 10));
    s.add(new ButtonInfo("Shuffle"      , 0x7689d827, 0, 9));
    s.add(new ButtonInfo("Rew"	        , 0x7689c03f, 0, 1));
    s.add(new ButtonInfo("Play"	        , 0x768910ef, 2, 2));
    s.add(new ButtonInfo("Fwd"	        , 0x7689a05f, 2, 1));
    s.add(new ButtonInfo("Add"	        , 0x7689609f, 0, 2));
    s.add(new ButtonInfo("Pause"	, 0x768920df, 1, 1));
    s.add(new ButtonInfo("Size"	        , 0x7689f807, 1, 10));
    s.add(new ButtonInfo("Sleep"	, 0x7689b847, 2, 9));
    s.add(new ButtonInfo("Up"	        , 0x7689e01f, 1, 2));
    s.add(new ButtonInfo("Left"	        , 0x7689096f, 0, 3));
    s.add(new ButtonInfo("Right"	, 0x7689d02f, 2, 3));
    s.add(new ButtonInfo("Down"	        , 0x7689b04f, 1, 4));
    s.add(new ButtonInfo("Brightness"	, 0x768904fb, 2, 10));
    s.add(new ButtonInfo("Repeat"	, 0x768938c7, 1, 9));
    s.add(new ButtonInfo("Power"	, 0x768940bf, 2, 0));
    s.add(new ButtonInfo("Vol-"	        , 0x768900ff, 0, 0));
    s.add(new ButtonInfo("Vol+"	        , 0x7689807f, 1, 0));
  }
}
