import java.util.*;
import java.io.*;

public class VFDCharFactory {

  private HashMap cache_;
  
  public VFDCharFactory() {
    cache_ = new HashMap(0xff);
  }
  
  VFDChar getCharForCode(int code) {
    VFDChar c = (VFDChar) cache_.get(new Integer(code));
    if (c != null)
      return c;
    else {
      // do something smart...
      return new VFDChar();
    }
  }

  public void addCharForCode(int code, VFDChar c) {
    cache_.put(new Integer(code), c);
  }

  public void loadFontFromFile(String f) 
    throws IOException {
    BufferedReader br = new BufferedReader(new FileReader(f));
    
    Vector currLines = new Vector(8);
    while (true) {
      String line = br.readLine();
      if (line == null) // EOF
	break;
      if (nonBlank(line))
	currLines.add(line);
      if (currLines.size() == 8) {
	tryToAddChar(currLines);
	currLines = new Vector(8);
      }
    }
  }

  private boolean nonBlank(String line) {
    // maybe it's a comment, check for '#' as first non-blank
    // character on line
    for (int i = 0; i < line.length(); i++) {
      char c = line.charAt(i);
      if (Character.isWhitespace(c))
	continue; // check next char
      else if (c == '#')
	return false;
      else 
	return true;
    }
    return false;
  }

  private void tryToAddChar(Vector v) {
    Util.assert(v.size() == 8);
    VFDChar vfdChar = new VFDChar();
    String s1 = (String) v.get(0);
    int code;
    try { code = Integer.parseInt(s1.trim(), 16); }
    catch (NumberFormatException x) { 
      Util.dbg("Bad character code `" + s1.trim() + "' in font file");
      return;
    }
    for (int y = 0; y < 7; y++) {
      String l = (String) v.get(1 + y);
      int x = 0;
      for (int i = 0; i < l.length(); i++) {
	char c = l.charAt(i);
	if (Character.isWhitespace(c))
	  continue;
	else if (c != '0')
	  vfdChar.setDot(x, y, true);
	x++;
      }
    }
    cache_.put(new Integer(code), vfdChar);
  }
}

